import sys
import datetime

PY2 = sys.version_info[0] == 2


def float_value(value):
    """convert a value to float"""
    ret = float(value)
    return ret


def date_value(value):
    """convert to data value accroding ods specification"""
    ret = "invalid"
    try:
        # catch strptime exceptions only
        if len(value) == 10:
            ret = datetime.datetime.strptime(
                value,
                "%Y-%m-%d")
            ret = ret.date()
        elif len(value) == 19:
            ret = datetime.datetime.strptime(
                value,
                "%Y-%m-%dT%H:%M:%S")
        elif len(value) > 19:
            ret = datetime.datetime.strptime(
                value[0:26],
                "%Y-%m-%dT%H:%M:%S.%f")
    except ValueError:
        pass
    if ret == "invalid":
        raise Exception("Bad date value %s" % value)
    return ret


def time_value(value):
    """convert to time value accroding the specification"""
    import re
    results = re.match('PT(\d+)H(\d+)M(\d+)S', value)
    if results and len(results.groups()) == 3:
        hour = int(results.group(1))
        minute = int(results.group(2))
        second = int(results.group(3))
        if hour < 24:
            ret = datetime.time(hour, minute, second)
        else:
            ret = datetime.timedelta(hours=hour,
                                     minutes=minute,
                                     seconds=second)
    else:
        ret = None
    return ret


def boolean_value(value):
    """get bolean value"""
    if value == "true":
        ret = True
    else:
        ret = False
    return ret


ODS_FORMAT_CONVERSION = {
    "float": float,
    "date": datetime.date,
    "time": datetime.time,
    'timedelta': datetime.timedelta,
    "boolean": bool,
    "percentage": float,
    "currency": float
}


ODS_WRITE_FORMAT_COVERSION = {
    float: "float",
    int: "float",
    str: "string",
    datetime.date: "date",
    datetime.time: "time",
    datetime.timedelta: "timedelta",
    bool: "boolean"
}

if PY2:
    ODS_WRITE_FORMAT_COVERSION[unicode] = "string"

VALUE_CONVERTERS = {
    "float": float_value,
    "date": date_value,
    "time": time_value,
    "timedelta": time_value,
    "boolean": boolean_value,
    "percentage": float_value
}


VALUE_TOKEN = {
    "float": "value",
    "date": "date-value",
    "time": "time-value",
    "boolean": "boolean-value",
    "percentage": "value",
    "currency": "value",
    "timedelta": "time-value"
}



"""
    pyexcel_odsr
    ~~~~~~~~~~~~~~~~~~~
    The lower level ods file format handler using messytables
    :copyright: (c) 2015-2017 by Onni Software Ltd & its contributors
    :license: New BSD License
"""
from pyexcel_io.plugins import IOPluginInfoChain
from pyexcel_io.io import get_data as read_data, isstream
from pyexcel_odsr._version import __version__, __author__  # flake8: noqa

__FILE_TYPE__ = 'ods'
__FILE_TYPE_FODS__ = 'fods'

IOPluginInfoChain(__name__).add_a_reader(
    relative_plugin_class_path='odsr.ODSBook',
    file_types=[__FILE_TYPE__],
    stream_type='binary'
).add_a_reader(
    relative_plugin_class_path='odsr.FODSBook',
    file_types=[__FILE_TYPE_FODS__],
    stream_type='binary'
)


def get_data(afile, file_type=None, **keywords):
    """standalone module function for reading module supported file type"""
    if isstream(afile) and file_type is None:
        file_type = __FILE_TYPE__
    return read_data(afile, file_type=file_type, **keywords)



"""
Copyright (c) 2012-2017 The Open Knowledge Foundation Ltd.

Permission is hereby granted, free of charge, to any person obtaining a copy of
this software and associated documentation files (the "Software"), to deal in
the Software without restriction, including without limitation the rights to
use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies
of the Software, and to permit persons to whom the Software is furnished to do
so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.
"""
import io
import re
import zipfile

from lxml import etree

from pyexcel_odsr.converter import VALUE_TOKEN, PY2

ODS_NAMESPACES_TAG_MATCH = re.compile(b"(<office:document-content[^>]*>)",
                                      re.MULTILINE)
ODS_TABLE_MATCH = re.compile(b".*?(<table:table.*?<\/.*?:table>).*?",
                             re.MULTILINE)
ODS_TABLE_NAME = re.compile(b".*?table:name=\"(.*?)\".*?")
ODS_ROW_MATCH = re.compile(b".*?(<table:table-row.*?<\/.*?:table-row>).*?",
                           re.MULTILINE)
ODS_DOCUMENT_CLOSE_TAG = b'</office:document-content>'
FODS_NAMESPACES_TAG_MATCH = re.compile(b"(<office:document[^>]*>)",
                                       re.DOTALL)
FODS_TABLE_MATCH = re.compile(b".*?(<table:table.*?<\/.*?:table>).*?",
                              re.DOTALL)
FODS_TABLE_NAME = re.compile(b".*?table:name=\"(.*?)\".*?")
FODS_ROW_MATCH = re.compile(b".*?(<table:table-row.*?<\/.*?:table-row>).*?",
                            re.DOTALL)
FODS_DOCUMENT_CLOSE_TAG = b'</office:document>'
NS_OPENDOCUMENT_PTTN = u"urn:oasis:names:tc:opendocument:xmlns:%s"
NS_CAL_PTTN = u"urn:org:documentfoundation:names:experimental:calc:xmlns:%s"
NS_OPENDOCUMENT_TABLE = NS_OPENDOCUMENT_PTTN % "table:1.0"
NS_OPENDOCUMENT_OFFICE = NS_OPENDOCUMENT_PTTN % "office:1.0"

TABLE_CELL = 'table-cell'
VALUE_TYPE = 'value-type'
COLUMN_REPEAT = 'number-columns-repeated'

DEFAULT_NAMESPACES = {
    "dc": u"http://purl.org/dc/elements/1.1/",
    "draw": NS_OPENDOCUMENT_PTTN % u"drawing:1.0",
    "number": NS_OPENDOCUMENT_PTTN % u"datastyle:1.0",
    "office": NS_OPENDOCUMENT_PTTN % u"office:1.0",
    "svg": NS_OPENDOCUMENT_PTTN % u"svg-compatible:1.0",
    "table": NS_OPENDOCUMENT_PTTN % u"table:1.0",
    "text": NS_OPENDOCUMENT_PTTN % u"text:1.0",
    "calcext": NS_CAL_PTTN % u"calcext:1.0",
}


class ODSTableSet(object):
    """
    A wrapper around ODS files. Because they are zipped and the info we want
    is in the zipped file as content.xml we must ensure that we either have
    a seekable object (local file) or that we retrieve all of the content from
    the remote URL.
    """

    def __init__(self, fileobj, window=None, **kw):
        '''Initialize the object.

        :param fileobj: may be a file path or a file-like object. Note the
        file-like object *must* be in binary mode and must be seekable (it will
        get passed to zipfile).

        As a specific tip: urllib2.urlopen returns a file-like object that is
        not in file-like mode while urllib.urlopen *does*!

        To get a seekable file you *cannot* use
        messytables.core.seekable_stream as it does not support the full seek
        functionality.
        '''
        if hasattr(fileobj, 'read'):
            # wrap in a StringIO so we do not have hassle with seeks and
            # binary etc (see notes to __init__ above)
            # TODO: rather wasteful if in fact fileobj comes from disk
            fileobj = io.BytesIO(fileobj.read())

        self.window = window

        zf = zipfile.ZipFile(fileobj).open("content.xml")
        self.content = zf.read()
        zf.close()
        self._table_matcher = ODS_TABLE_MATCH
        self._document_close_tag = ODS_DOCUMENT_CLOSE_TAG
        self._namespace_tag_matcher = ODS_NAMESPACES_TAG_MATCH
        self._row_set_cls = ODSRowSet

    def make_tables(self):
        """
            Return the sheets in the workbook.

            A regex is used for this to avoid having to:

            1. load large the entire file into memory, or
            2. SAX parse the file more than once
        """
        namespace_tags = self._get_namespace_tags()
        sheets = [m.groups(0)[0]
                  for m in self._table_matcher.finditer(self.content)]
        return [self._row_set_cls(sheet, self.window, namespace_tags)
                for sheet in sheets]

    def _get_namespace_tags(self):
        match = re.search(self._namespace_tag_matcher, self.content)
        assert match
        tag_open = match.groups()[0]
        tag_close = self._document_close_tag
        return tag_open, tag_close


class ODSRowSet(object):
    """ ODS support for a single sheet in the ODS workbook. Unlike
    the CSV row set this is not a streaming operation. """

    def __init__(self, sheet, window=None, namespace_tags=None):
        self.sheet = sheet

        self.name = "Unknown"
        m = ODS_TABLE_NAME.match(self.sheet)
        if m:
            self.name = m.groups(0)[0]
            if not PY2 and isinstance(self.name, bytes):
                self.name = self.name.decode('utf-8')

        self.window = window or 1000

        # We must wrap the XML fragments in a valid header otherwise iterparse
        # will explode with certain (undefined) versions of libxml2. The
        # namespaces are in the ODS file, and change with the libreoffice
        # version saving it, so get them from the ODS file if possible. The
        # default namespaces are an option to preserve backwards compatibility
        # of ODSRowSet.
        if namespace_tags:
            self.namespace_tags = namespace_tags
        else:
            namespaces = DEFAULT_NAMESPACES

            ods_header = u"<wrapper {0}>"\
                .format(" ".join('xmlns:{0}="{1}"'.format(k, v)
                        for k, v in namespaces.iteritems())).encode('utf-8')
            ods_footer = u"</wrapper>".encode('utf-8')
            self.namespace_tags = (ods_header, ods_footer)

        self._row_matcher = ODS_ROW_MATCH

    def raw(self, sample=False):
        """ Iterate over all rows in this sheet. """
        rows = self._row_matcher.findall(self.sheet)

        for row in rows:
            row_data = []

            block = self.namespace_tags[0] + row + self.namespace_tags[1]
            partial = io.BytesIO(block)

            for action, element in etree.iterparse(partial, ('end',)):
                if element.tag != _tag(NS_OPENDOCUMENT_TABLE, TABLE_CELL):
                    continue

                cell = _read_cell(element)
                repeat = element.attrib.get(
                    _tag(NS_OPENDOCUMENT_TABLE, COLUMN_REPEAT))

                if repeat:
                    number_of_repeat = int(repeat)
                    row_data += [cell] * number_of_repeat
                else:
                    row_data.append(cell)

            del partial
            yield row_data
        del rows


class FODSTableSet(ODSTableSet):
    """
    A wrapper around ODS files. Because they are zipped and the info we want
    is in the zipped file as content.xml we must ensure that we either have
    a seekable object (local file) or that we retrieve all of the content from
    the remote URL.
    """

    def __init__(self, fileobj, window=None, **kw):
        '''Initialize the object.

        :param fileobj: may be a file path or a file-like object. Note the
        file-like object *must* be in binary mode and must be seekable (it will
        get passed to zipfile).

        As a specific tip: urllib2.urlopen returns a file-like object that is
        not in file-like mode while urllib.urlopen *does*!

        To get a seekable file you *cannot* use
        messytables.core.seekable_stream as it does not support the full seek
        functionality.
        '''
        if hasattr(fileobj, 'read'):
            self.content = fileobj.read()
        else:
            with open(fileobj, 'rb') as f:
                self.content = f.read()

        self.window = window

        self._table_matcher = FODS_TABLE_MATCH
        self._document_close_tag = FODS_DOCUMENT_CLOSE_TAG
        self._namespace_tag_matcher = FODS_NAMESPACES_TAG_MATCH
        self._row_set_cls = FODSRowSet


class FODSRowSet(ODSRowSet):
    """ ODS support for a single sheet in the ODS workbook. Unlike
    the CSV row set this is not a streaming operation. """

    def __init__(self, sheet, window=None, namespace_tags=None):
        super(FODSRowSet, self).__init__(sheet, window, namespace_tags)
        self._row_matcher = FODS_ROW_MATCH


def _read_cell(element):
    cell_type = element.attrib.get(_tag(NS_OPENDOCUMENT_OFFICE, VALUE_TYPE))
    value_token = VALUE_TOKEN.get(cell_type, 'value')
    if cell_type == 'string':
        cell = _read_text_cell(element)
    elif cell_type == 'currency':
        value = element.attrib.get(_tag(NS_OPENDOCUMENT_OFFICE, value_token))
        currency = element.attrib.get(_tag(NS_OPENDOCUMENT_OFFICE, 'currency'))
        cell = (value + ' ' + currency, 'currency')
    elif cell_type is not None:
        value = element.attrib.get(_tag(NS_OPENDOCUMENT_OFFICE, value_token))
        cell = (value, cell_type)
    else:
        cell = ('', 'string')
    return cell


def _read_text_cell(element):
    children = element.getchildren()
    text_content = []
    for child in children:
        text = "".join([x for x in child.itertext()])
        text_content.append(text)
    if len(text_content) > 0:
        cell_value = '\n'.join(text_content)
    else:
        cell_value = ''
    return (cell_value, 'string')


def _tag(namespace, tag):
    return '{%s}%s' % (namespace, tag)



"""
    pyexcel_odsr.odsr
    ~~~~~~~~~~~~~~~~~~~
    The lower level ods file format handler using messytables

    :copyright: (c) 2015-2017 by Onni Software Ltd & its contributors
    :license: New BSD License
"""
import math
from io import UnsupportedOperation

from pyexcel_io.book import BookReader
from pyexcel_io.sheet import SheetReader
from pyexcel_io._compact import OrderedDict, BytesIO

from pyexcel_odsr.messyods import ODSTableSet, FODSTableSet
import pyexcel_odsr.converter as converter


class ODSSheet(SheetReader):
    """native ods sheet"""
    def __init__(self, sheet, auto_detect_int=True,
                 auto_detect_float=True,
                 auto_detect_datetime=True,
                 **keywords):
        SheetReader.__init__(self, sheet, **keywords)
        self.__auto_detect_int = auto_detect_int
        self.__auto_detect_float = auto_detect_float
        self.__auto_detect_datetime = auto_detect_datetime

    @property
    def name(self):
        return self._native_sheet.name

    def row_iterator(self):
        return self._native_sheet.raw()

    def column_iterator(self, row):
        for cell in row:
            yield self.__convert_cell(cell)

    def __convert_cell(self, cell):
        ret = None
        if cell[1] in converter.VALUE_CONVERTERS:
            n_value = converter.VALUE_CONVERTERS[cell[1]](cell[0])
            if cell[1] == 'float' and self.__auto_detect_int:
                if is_integer_ok_for_xl_float(n_value):
                    n_value = int(n_value)
            ret = n_value
        else:
            ret = cell[0]
        return ret


class ODSBook(BookReader):
    """read ods book"""
    def open(self, file_name, **keywords):
        """open ods file"""
        BookReader.open(self, file_name, **keywords)
        self._load_from_file()

    def open_stream(self, file_stream, **keywords):
        """open ods file stream"""
        if not hasattr(file_stream, 'seek'):
            # python 2
            # Hei zipfile in odfpy would do a seek
            # but stream from urlib cannot do seek
            file_stream = BytesIO(file_stream.read())
        try:
            file_stream.seek(0)
        except UnsupportedOperation:
            # python 3
            file_stream = BytesIO(file_stream.read())
        BookReader.open_stream(self, file_stream, **keywords)
        self._load_from_memory()

    def read_sheet_by_name(self, sheet_name):
        """read a named sheet"""
        tables = self._native_book.make_tables()
        rets = [table for table in tables
                if table.name == sheet_name]
        if len(rets) == 0:
            raise ValueError("%s cannot be found" % sheet_name)
        else:
            return self.read_sheet(rets[0])

    def read_sheet_by_index(self, sheet_index):
        """read a sheet at a specified index"""
        tables = self._native_book.make_tables()
        length = len(tables)
        if sheet_index < length:
            return self.read_sheet(tables[sheet_index])
        else:
            raise IndexError("Index %d of out bound %d" % (
                sheet_index, length))

    def read_all(self):
        """read all sheets"""
        result = OrderedDict()
        for sheet in self._native_book.make_tables():
            ods_sheet = ODSSheet(sheet, **self._keywords)
            result[ods_sheet.name] = ods_sheet.to_array()

        return result

    def read_sheet(self, native_sheet):
        """read one native sheet"""
        sheet = ODSSheet(native_sheet, **self._keywords)
        return {sheet.name: sheet.to_array()}

    def _load_from_memory(self):
        self._native_book = ODSTableSet(self._file_stream)

    def _load_from_file(self):
        self._native_book = ODSTableSet(self._file_name)


class FODSBook(ODSBook):
    """read fods book"""
    def _load_from_file(self):
        self._native_book = FODSTableSet(self._file_name)

    def _load_from_memory(self):
        self._native_book = FODSTableSet(self._file_stream)


def is_integer_ok_for_xl_float(value):
    """check if a float had zero value in digits"""
    return value == math.floor(value)
